@@ -7,12 +7,19 @@
  */
 #include "eim.h"
 #include <cstdarg>
+#include <fcitx-utils/keysymgen.h>
+#include <fcitx-utils/log.h>
 #include <fcitx-utils/utf8.h>
+#include <fcitx/candidatelist.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputpanel.h>
 #include <fcitx/statusarea.h>
 #include <fcitx/text.h>
+#include <fcitx/userinterface.h>
 #include <fcitx/userinterfacemanager.h>
+#include <memory>
+#include <utility>
 
 FCITX_DEFINE_LOG_CATEGORY(chewing_log, "chewing");
 #define CHEWING_DEBUG() FCITX_LOGC(chewing_log, Debug)
@@ -23,12 +30,7 @@ namespace {
 
 constexpr int CHEWING_MAX_LEN = 18;
 
-std::string builtin_keymaps[] = {
-    "KB_DEFAULT",     "KB_HSU",          "KB_IBM",    "KB_GIN_YEIH",
-    "KB_ET",          "KB_ET26",         "KB_DVORAK", "KB_DVORAK_HSU",
-    "KB_DACHEN_CP26", "KB_HANYU_PINYIN", "KB_CARPALX"};
-
-static const char *builtin_selectkeys[] = {
+const char *builtin_selectkeys[] = {
     "1234567890", "asdfghjkl;", "asdfzxcv89", "asdfjkl789",
     "aoeuhtn789", "1234qweras", "dstnaeo789",
 };
@@ -37,18 +39,33 @@ static_assert(FCITX_ARRAY_SIZE(builtin_selectkeys) ==
                   ChewingSelectionKeyI18NAnnotation::enumLength,
               "Enum mismatch");
 
+#define DEFINE_SAFE_CHEWING_STRING_GETTER(NAME)                                \
+    static inline std::string safeChewing_##NAME##_String(                     \
+        ChewingContext *ctx) {                                                 \
+        if (chewing_##NAME##_Check(ctx)) {                                     \
+            return chewing_##NAME##_String_static(ctx);                        \
+        }                                                                      \
+        return "";                                                             \
+    }
+
+DEFINE_SAFE_CHEWING_STRING_GETTER(aux);
+DEFINE_SAFE_CHEWING_STRING_GETTER(buffer);
+DEFINE_SAFE_CHEWING_STRING_GETTER(bopomofo);
+DEFINE_SAFE_CHEWING_STRING_GETTER(commit);
+
 class ChewingCandidateWord : public CandidateWord {
 public:
     ChewingCandidateWord(ChewingEngine *engine, std::string str, int index)
-        : CandidateWord(Text(str)), engine_(engine), index_(index) {}
+        : CandidateWord(Text(std::move(str))), engine_(engine), index_(index) {}
 
     void select(InputContext *inputContext) const override {
-        auto pageSize = engine_->instance()->globalConfig().defaultPageSize();
-        auto ctx = engine_->context();
+        auto *ctx = engine_->context();
+        auto pageSize = chewing_get_candPerPage(ctx);
         int page = index_ / pageSize + chewing_cand_CurrentPage(ctx);
         int off = index_ % pageSize;
-        if (page < 0 || page >= chewing_cand_TotalPage(ctx))
+        if (page < 0 || page >= chewing_cand_TotalPage(ctx)) {
             return;
+        }
         int lastPage = chewing_cand_CurrentPage(ctx);
         while (page != chewing_cand_CurrentPage(ctx)) {
             if (page < chewing_cand_CurrentPage(ctx)) {
@@ -66,17 +83,14 @@ public:
         chewing_handle_Default(ctx, builtin_selectkeys[static_cast<int>(
                                         *engine_->config().SelectionKey)][off]);
 
-        if (chewing_keystroke_CheckAbsorb(ctx)) {
-            engine_->updateUI(inputContext);
-        } else if (chewing_keystroke_CheckIgnore(ctx)) {
+        if (chewing_keystroke_CheckIgnore(ctx)) {
             return;
-        } else if (chewing_commit_Check(ctx)) {
-            UniqueCPtr<char, chewing_free> str(chewing_commit_String(ctx));
-            inputContext->commitString(str.get());
-            engine_->updateUI(inputContext);
-        } else {
-            engine_->updateUI(inputContext);
         }
+
+        if (chewing_commit_Check(ctx)) {
+            inputContext->commitString(safeChewing_commit_String(ctx));
+        }
+        engine_->updateUI(inputContext);
     }
 
 private:
@@ -85,35 +99,17 @@ private:
 };
 
 class ChewingCandidateList : public CandidateList,
-                             public PageableCandidateList {
+                             public PageableCandidateList,
+                             public CursorMovableCandidateList,
+                             public CursorModifiableCandidateList {
 public:
     ChewingCandidateList(ChewingEngine *engine, InputContext *ic)
         : engine_(engine), ic_(ic) {
-        auto ctx = engine_->context();
         setPageable(this);
-        int index = 0;
-        // get candidate word
-        int pageSize = chewing_get_candPerPage(ctx);
-        chewing_cand_Enumerate(ctx);
-        while (chewing_cand_hasNext(ctx) && index < pageSize) {
-            UniqueCPtr<char, chewing_free> str(chewing_cand_String(ctx));
-            candidateWords_.emplace_back(std::make_unique<ChewingCandidateWord>(
-                engine_, str.get(), index));
-            if (index < 10) {
-                const char label[] = {
-                    builtin_selectkeys[static_cast<int>(
-                        *engine_->config().SelectionKey)][index],
-                    '.', '\0'};
-                labels_.emplace_back(label);
-            } else {
-                labels_.emplace_back();
-            }
-            index++;
-        }
+        setCursorMovable(this);
+        setCursorModifiable(this);
 
-        hasPrev_ = chewing_cand_CurrentPage(ctx) > 0;
-        hasNext_ =
-            chewing_cand_CurrentPage(ctx) + 1 < chewing_cand_TotalPage(ctx);
+        fillCandidate();
     }
 
     const Text &label(int idx) const override {
@@ -129,61 +125,157 @@ public:
         return *candidateWords_[idx];
     }
 
-    int size() const override { return candidateWords_.size(); }
-    int cursorIndex() const override { return -1; }
-    CandidateLayoutHint layoutHint() const override {
-        return *engine_->config().CandidateLayout;
+    void fillCandidate() {
+        auto *ctx = engine_->context();
+        candidateWords_.clear();
+        labels_.clear();
+        cursor_ = 0;
+
+        int index = 0;
+        // get candidate word
+        int pageSize = chewing_cand_ChoicePerPage(ctx);
+        if (pageSize <= 0) {
+            return;
+        }
+        chewing_cand_Enumerate(ctx);
+        while (chewing_cand_hasNext(ctx) && index < pageSize) {
+            candidateWords_.emplace_back(std::make_unique<ChewingCandidateWord>(
+                engine_, chewing_cand_String_static(ctx), index));
+            if (index < 10) {
+                const char label[] = {
+                    builtin_selectkeys[static_cast<int>(
+                        *engine_->config().SelectionKey)][index],
+                    '.', '\0'};
+                labels_.emplace_back(label);
+            } else {
+                labels_.emplace_back();
+            }
+            index++;
+        }
     }
 
-    // Need for paging
-    bool hasPrev() const override { return hasPrev_; }
-    bool hasNext() const override { return hasNext_; }
+    int size() const override { return candidateWords_.size(); }
+    int cursorIndex() const override {
+        if (empty() || !*engine_->config().selectCandidateWithArrowKey) {
+            return -1;
+        }
+        return cursor_;
+    }
+    CandidateLayoutHint layoutHint() const override {
+        switch (*engine_->config().CandidateLayout) {
+        case ChewingCandidateLayout::Horizontal:
+            return CandidateLayoutHint::Horizontal;
+        case ChewingCandidateLayout::Vertical:
+            return CandidateLayoutHint::Vertical;
+        }
+        return CandidateLayoutHint::Horizontal;
+    }
+
+    // Need for paging, allow rotating pages.
+    bool hasPrev() const override { return true; }
+    bool hasNext() const override { return true; }
     void prev() override { paging(true); }
     void next() override { paging(false); }
 
     bool usedNextBefore() const override { return true; }
 
+    void prevCandidate() override {
+        if (cursor_ == 0) {
+            prev();
+            // We do not reset cursor to the last on purpose.
+            // This way it could be easily to change the range of current word.
+        } else {
+            cursor_ -= 1;
+            ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
+        }
+    }
+
+    void nextCandidate() override {
+        if (cursor_ + 1 == size()) {
+            next();
+            cursor_ = 0;
+        } else {
+            cursor_ += 1;
+            ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
+        }
+    }
+
+    void setCursorIndex(int cursor) override {
+        if (cursor < 0 || cursor >= size()) {
+            return;
+        }
+        cursor_ = cursor;
+    }
+
 private:
     void paging(bool prev) {
         if (candidateWords_.empty()) {
             return;
         }
 
-        auto ctx = engine_->context();
+        auto *ctx = engine_->context();
+        const int currentPage = chewing_cand_CurrentPage(ctx);
         if (prev) {
-            chewing_handle_Left(ctx);
+            const int hasNext = chewing_cand_list_has_next(ctx);
+            const int hasPrev = chewing_cand_list_has_prev(ctx);
+            if ((currentPage == 0) && (hasNext == 1 || hasPrev == 1)) {
+                chewing_handle_Down(ctx);
+            } else {
+                chewing_handle_PageUp(ctx);
+            }
         } else {
-            chewing_handle_Right(ctx);
+            const int totalPage = chewing_cand_TotalPage(ctx);
+            if (currentPage == totalPage - 1) {
+                chewing_handle_Down(ctx);
+            } else {
+                chewing_handle_PageDown(ctx);
+            }
         }
 
         if (chewing_keystroke_CheckAbsorb(ctx)) {
-            engine_->updateUI(ic_);
+            fillCandidate();
+            engine_->updatePreedit(ic_);
+            ic_->updatePreedit();
+            ic_->updateUserInterface(UserInterfaceComponent::InputPanel);
         }
     }
 
-    bool hasPrev_ = false;
-    bool hasNext_ = false;
     ChewingEngine *engine_;
     InputContext *ic_;
     std::vector<std::unique_ptr<ChewingCandidateWord>> candidateWords_;
     std::vector<Text> labels_;
+    int cursor_ = 0;
 };
 
-void logger(void *, int, const char *fmt, ...) {
-    va_list ap;
-    va_start(ap, fmt);
-    vfprintf(stderr, fmt, ap);
-    va_end(ap);
+void logger(void * /*context*/, int /*level*/, const char *fmt, ...) {
+    if (!chewing_log().checkLogLevel(Debug)) {
+        return;
+    }
+    std::va_list argp;
+    va_start(argp, fmt);
+    char onechar[1];
+    int len = std::vsnprintf(onechar, 1, fmt, argp);
+    va_end(argp);
+    if (len < 1) {
+        return;
+    }
+    std::vector<char> buf;
+    buf.resize(len + 1);
+    buf.back() = 0;
+    va_start(argp, fmt);
+    std::vsnprintf(buf.data(), len, fmt, argp);
+    va_end(argp);
+    CHEWING_DEBUG() << buf.data();
 }
 
 } // namespace
 
 ChewingContext *getChewingContext() {
-    const auto &sp = fcitx::StandardPath::global();
+    const auto &sp = StandardPath::global();
     std::string dictData =
-        sp.locate(fcitx::StandardPath::Type::Data, "libchewing/dictionary.dat");
+        sp.locate(StandardPath::Type::Data, "libchewing/dictionary.dat");
     if (!dictData.empty()) {
-        std::string sysPath = fcitx::fs::dirName(dictData);
+        std::string sysPath = fs::dirName(dictData);
         return chewing_new2(sysPath.c_str(), nullptr, nullptr, nullptr);
     }
     return chewing_new();
@@ -192,9 +284,7 @@ ChewingContext *getChewingContext() {
 ChewingEngine::ChewingEngine(Instance *instance)
     : instance_(instance), context_(getChewingContext()) {
     chewing_set_maxChiSymbolLen(context_.get(), CHEWING_MAX_LEN);
-    if (chewing_log().checkLogLevel(Debug)) {
-        chewing_set_logger(context_.get(), logger, nullptr);
-    }
+    chewing_set_logger(context_.get(), logger, nullptr);
     reloadConfig();
 }
 
@@ -208,17 +298,18 @@ void ChewingEngine::reloadConfig() {
 void ChewingEngine::populateConfig() {
     ChewingContext *ctx = context_.get();
 
+    CHEWING_DEBUG() << "Set layout to: "
+                    << builtin_keymaps[static_cast<int>(*config_.Layout)];
     chewing_set_KBType(
-        ctx, chewing_KBStr2Num(
-                 builtin_keymaps[static_cast<int>(*config_.Layout)].data()));
+        ctx,
+        chewing_KBStr2Num(builtin_keymaps[static_cast<int>(*config_.Layout)]));
 
     chewing_set_ChiEngMode(ctx, CHINESE_MODE);
 
     int selkey[10];
-    int i = 0;
-    for (i = 0; i < 10; i++) {
-        selkey[i] =
-            builtin_selectkeys[static_cast<int>(*config_.SelectionKey)][i];
+    for (size_t i = 0; i < 10; i++) {
+        selkey[i] = static_cast<unsigned char>(
+            builtin_selectkeys[static_cast<int>(*config_.SelectionKey)][i]);
     }
 
     chewing_set_selKey(ctx, selkey, 10);
@@ -230,19 +321,22 @@ void ChewingEngine::populateConfig() {
     chewing_set_escCleanAllBuf(ctx, 1);
 }
 
-void ChewingEngine::reset(const InputMethodEntry &, InputContextEvent &event) {
+void ChewingEngine::reset(const InputMethodEntry & /*entry*/,
+                          InputContextEvent &event) {
     doReset(event);
 }
 
 void ChewingEngine::doReset(InputContextEvent &event) {
     ChewingContext *ctx = context_.get();
-    chewing_handle_Esc(ctx);
+    chewing_cand_close(ctx);
+    chewing_clean_preedit_buf(ctx);
+    chewing_clean_bopomofo_buf(ctx);
     updateUI(event.inputContext());
 }
 
 void ChewingEngine::save() {}
 
-void ChewingEngine::activate(const InputMethodEntry &,
+void ChewingEngine::activate(const InputMethodEntry & /*entry*/,
                              InputContextEvent &event) {
     // Request chttrans.
     // Fullwidth is not required for chewing.
@@ -252,41 +346,108 @@ void ChewingEngine::activate(const InputMethodEntry &,
             instance_->userInterfaceManager().lookupAction("chttrans")) {
         inputContext->statusArea().addAction(StatusGroup::InputMethod, action);
     }
+    auto *ic = event.inputContext();
+    if (!ic_.isNull() && ic_.get() != ic) {
+        doReset(event);
+    }
+    ic_ = ic->watch();
 }
 
 void ChewingEngine::deactivate(const InputMethodEntry &entry,
                                InputContextEvent &event) {
-    if (event.type() == EventType::InputContextFocusOut ||
-        event.type() == EventType::InputContextSwitchInputMethod) {
+    if (event.type() == EventType::InputContextSwitchInputMethod) {
         flushBuffer(event);
     } else {
         reset(entry, event);
     }
 }
 
-void ChewingEngine::keyEvent(const InputMethodEntry &entry,
-                             KeyEvent &keyEvent) {
-    auto ctx = context_.get();
-    if (keyEvent.isRelease()) {
-        return;
+bool ChewingEngine::handleCandidateKeyEvent(const KeyEvent &keyEvent) {
+    auto *ic = keyEvent.inputContext();
+    auto candidateList = std::dynamic_pointer_cast<ChewingCandidateList>(
+        ic->inputPanel().candidateList());
+    if (!candidateList) {
+        return false;
     }
 
-    chewing_set_easySymbolInput(ctx, 0);
-    CHEWING_DEBUG() << "KeyEvent: " << keyEvent.key().toString();
-    auto ic = keyEvent.inputContext();
     const KeyList keypadKeys{Key{FcitxKey_KP_1}, Key{FcitxKey_KP_2},
                              Key{FcitxKey_KP_3}, Key{FcitxKey_KP_4},
                              Key{FcitxKey_KP_5}, Key{FcitxKey_KP_6},
                              Key{FcitxKey_KP_7}, Key{FcitxKey_KP_8},
                              Key{FcitxKey_KP_9}, Key{FcitxKey_KP_0}};
-    if (*config_.UseKeypadAsSelectionKey && ic->inputPanel().candidateList()) {
+    if (*config_.UseKeypadAsSelectionKey) {
         if (int index = keyEvent.key().keyListIndex(keypadKeys);
-            index >= 0 && index < ic->inputPanel().candidateList()->size()) {
-            ic->inputPanel().candidateList()->candidate(index).select(ic);
-            return keyEvent.filterAndAccept();
+            index >= 0 && index < candidateList->size()) {
+            candidateList->candidate(index).select(ic);
+            return true;
         }
     }
 
+    if (!*config_.selectCandidateWithArrowKey) {
+        return false;
+    }
+
+    if (keyEvent.key().check(FcitxKey_Right)) {
+        if (*config_.CandidateLayout == ChewingCandidateLayout::Horizontal) {
+            candidateList->nextCandidate();
+        } else {
+            candidateList->next();
+        }
+        return true;
+    }
+    if (keyEvent.key().check(FcitxKey_Left)) {
+        if (*config_.CandidateLayout == ChewingCandidateLayout::Horizontal) {
+            candidateList->prevCandidate();
+        } else {
+            candidateList->prev();
+        }
+        return true;
+    }
+    if (keyEvent.key().check(FcitxKey_Up)) {
+        if (*config_.CandidateLayout == ChewingCandidateLayout::Horizontal) {
+            candidateList->prev();
+        } else {
+            candidateList->prevCandidate();
+        }
+        return true;
+    }
+    if (keyEvent.key().check(FcitxKey_Down)) {
+        if (*config_.CandidateLayout == ChewingCandidateLayout::Horizontal) {
+            candidateList->next();
+        } else {
+            candidateList->nextCandidate();
+        }
+        return true;
+    }
+    if (keyEvent.key().check(FcitxKey_Return)) {
+        if (int index = candidateList->cursorIndex();
+            index >= 0 && index < candidateList->size()) {
+            candidateList->candidate(index).select(ic);
+        }
+        return true;
+    }
+    if (keyEvent.key().check(FcitxKey_space)) {
+        candidateList->next();
+        return true;
+    }
+    return false;
+}
+
+void ChewingEngine::keyEvent(const InputMethodEntry &entry,
+                             KeyEvent &keyEvent) {
+    if (keyEvent.isRelease()) {
+        return;
+    }
+    auto *ctx = context_.get();
+    auto *ic = keyEvent.inputContext();
+
+    chewing_set_easySymbolInput(ctx, 0);
+    CHEWING_DEBUG() << "KeyEvent: " << keyEvent.key().toString();
+
+    if (handleCandidateKeyEvent(keyEvent)) {
+        return keyEvent.filterAndAccept();
+    }
+
     if (keyEvent.key().check(FcitxKey_space)) {
         chewing_handle_Space(ctx);
     } else if (keyEvent.key().check(FcitxKey_Tab)) {
@@ -297,34 +458,36 @@ void ChewingEngine::keyEvent(const InputMethodEntry &entry,
         }
         int scan_code = keyEvent.key().sym() & 0xff;
         if (*config_.Layout == ChewingLayout::HanYuPinYin) {
-            const char *zuin_str = chewing_bopomofo_String_static(ctx);
-            // Workaround a bug in libchewing fixed in 2017 but never has stable
-            // release.
-            if (std::string_view(zuin_str).size() >= 9) {
+            auto zuin = safeChewing_bopomofo_String(ctx);
+            // Workaround a bug in libchewing fixed in 2017 but never has
+            // stable release.
+            if (zuin.size() >= 9) {
                 return keyEvent.filterAndAccept();
             }
         }
         chewing_handle_Default(ctx, scan_code);
         chewing_set_easySymbolInput(ctx, 0);
     } else if (keyEvent.key().check(FcitxKey_BackSpace)) {
-        const char *zuin_str = chewing_bopomofo_String_static(ctx);
-        if (chewing_buffer_Len(ctx) == 0 && !zuin_str[0]) {
+        if ((chewing_buffer_Check(ctx)) == 0 &&
+            (chewing_bopomofo_Check(ctx) == 0)) {
             return;
         }
         chewing_handle_Backspace(ctx);
-        if (chewing_buffer_Len(ctx) == 0 && !zuin_str[0]) {
+        if ((chewing_buffer_Check(ctx)) == 0 &&
+            (chewing_bopomofo_Check(ctx) == 0)) {
             keyEvent.filterAndAccept();
             return reset(entry, keyEvent);
         }
     } else if (keyEvent.key().check(FcitxKey_Escape)) {
         chewing_handle_Esc(ctx);
     } else if (keyEvent.key().check(FcitxKey_Delete)) {
-        const char *zuin_str = chewing_bopomofo_String_static(ctx);
-        if (chewing_buffer_Len(ctx) == 0 && !zuin_str[0]) {
+        if ((chewing_buffer_Check(ctx)) == 0 &&
+            (chewing_bopomofo_Check(ctx) == 0)) {
             return;
         }
         chewing_handle_Del(ctx);
-        if (chewing_buffer_Len(ctx) == 0 && !zuin_str[0]) {
+        if ((chewing_buffer_Check(ctx)) == 0 &&
+            (chewing_bopomofo_Check(ctx) == 0)) {
             keyEvent.filterAndAccept();
             return reset(entry, keyEvent);
         }
@@ -360,27 +523,25 @@ void ChewingEngine::keyEvent(const InputMethodEntry &entry,
         return;
     }
 
+    if (chewing_keystroke_CheckIgnore(ctx)) {
+        return;
+    }
     if (chewing_keystroke_CheckAbsorb(ctx)) {
         keyEvent.filterAndAccept();
-        return updateUI(ic);
-    } else if (chewing_keystroke_CheckIgnore(ctx)) {
-        return;
-    } else if (chewing_commit_Check(ctx)) {
-        keyEvent.filterAndAccept();
-        UniqueCPtr<char, chewing_free> str(chewing_commit_String(ctx));
-        ic->commitString(str.get());
-        return updateUI(ic);
-    } else {
-        keyEvent.filterAndAccept();
-        return updateUI(ic);
     }
+    if (chewing_commit_Check(ctx)) {
+        keyEvent.filterAndAccept();
+        ic->commitString(safeChewing_commit_String(ctx));
+    }
+    return updateUI(ic);
 }
 
-void ChewingEngine::filterKey(const InputMethodEntry &, KeyEvent &keyEvent) {
+void ChewingEngine::filterKey(const InputMethodEntry & /*entry*/,
+                              KeyEvent &keyEvent) {
     if (keyEvent.isRelease()) {
         return;
     }
-    auto ic = keyEvent.inputContext();
+    auto *ic = keyEvent.inputContext();
     if (ic->inputPanel().candidateList() &&
         (keyEvent.key().isSimple() || keyEvent.key().isCursorMove() ||
          keyEvent.key().check(FcitxKey_space, KeyState::Shift) ||
@@ -398,32 +559,20 @@ void ChewingEngine::filterKey(const InputMethodEntry &, KeyEvent &keyEvent) {
     }
 }
 
-void ChewingEngine::updateUI(InputContext *ic) {
-    CHEWING_DEBUG() << "updateUI";
+void ChewingEngine::updatePreeditImpl(InputContext *ic) {
+    ic->inputPanel().setClientPreedit(Text());
+    ic->inputPanel().setPreedit(Text());
+    ic->inputPanel().setAuxDown(Text());
+
     ChewingContext *ctx = context_.get();
+    std::string buffer = safeChewing_buffer_String(ctx);
+    std::string_view text = buffer;
+    std::string zuin = safeChewing_bopomofo_String(ctx);
 
-    // clean up window asap
-    ic->inputPanel().reset();
-    ic->updateUserInterface(UserInterfaceComponent::InputPanel);
-
-    UniqueCPtr<char, chewing_free> buf_str(chewing_buffer_String(ctx));
-    const char *zuin_str = chewing_bopomofo_String_static(ctx);
-
-    std::string text = buf_str.get();
-    std::string_view zuin = zuin_str;
     CHEWING_DEBUG() << "Text: " << text << " Zuin: " << zuin;
-    /* if not check done, so there is candidate word */
-    if (!chewing_cand_CheckDone(ctx)) {
-        ic->inputPanel().setCandidateList(
-            std::make_unique<ChewingCandidateList>(this, ic));
-        if (!ic->inputPanel().candidateList()->size()) {
-            ic->inputPanel().setCandidateList(nullptr);
-        }
-    }
 
     /* there is nothing */
-    if (zuin.empty() && text.empty() && !ic->inputPanel().candidateList()) {
-        ic->updatePreedit();
+    if (zuin.empty() && text.empty()) {
         return;
     }
 
@@ -445,14 +594,12 @@ void ChewingEngine::updateUI(InputContext *ic) {
     preedit.setCursor(rcur);
 
     // insert zuin in the middle
-    preedit.append(text.substr(0, rcur), format);
-    preedit.append(std::string(zuin), {TextFormatFlag::HighLight, format});
-    preedit.append(text.substr(rcur), format);
+    preedit.append(std::string(text.substr(0, rcur)), format);
+    preedit.append(std::move(zuin), {TextFormatFlag::HighLight, format});
+    preedit.append(std::string(text.substr(rcur)), format);
 
-    if (chewing_aux_Check(ctx)) {
-        const char *aux_str = chewing_aux_String_static(ctx);
-        std::string aux = aux_str;
-        ic->inputPanel().setAuxDown(Text(aux));
+    if (auto aux = safeChewing_aux_String(ctx); !aux.empty()) {
+        ic->inputPanel().setAuxDown(Text(std::move(aux)));
     }
 
     if (useClientPreedit) {
@@ -460,29 +607,53 @@ void ChewingEngine::updateUI(InputContext *ic) {
     } else {
         ic->inputPanel().setPreedit(preedit);
     }
+}
 
+void ChewingEngine::updatePreedit(InputContext *ic) {
+    updatePreeditImpl(ic);
     ic->updatePreedit();
+}
+
+void ChewingEngine::updateUI(InputContext *ic) {
+    CHEWING_DEBUG() << "updateUI";
+    // clean up window asap
+    ic->inputPanel().reset();
+    ic->inputPanel().setCandidateList(
+        std::make_unique<ChewingCandidateList>(this, ic));
+    if (ic->inputPanel().candidateList()->empty()) {
+        ic->inputPanel().setCandidateList(nullptr);
+    }
+
+    updatePreedit(ic);
     ic->updateUserInterface(UserInterfaceComponent::InputPanel);
 }
 
 void ChewingEngine::flushBuffer(InputContextEvent &event) {
-    auto ctx = context_.get();
-    // This check is because we ask the client to do the focus out commit.
-    if (event.type() != EventType::InputContextFocusOut) {
-        chewing_handle_Enter(ctx);
-        if (chewing_commit_Check(ctx)) {
-            UniqueCPtr<char, chewing_free> str(chewing_commit_String(ctx));
-            event.inputContext()->commitString(str.get());
-        }
-        UniqueCPtr<char, chewing_free> buf_str(chewing_buffer_String(ctx));
-        const char *zuin_str = chewing_bopomofo_String_static(ctx);
-        std::string text = buf_str.get();
-        std::string zuin = zuin_str;
-        text += zuin;
-        if (!text.empty()) {
-            event.inputContext()->commitString(text);
+    auto *ctx = context_.get();
+    std::string text;
+    if (*config_.switchInputMethodBehavior ==
+            SwitchInputMethodBehavior::CommitPreedit ||
+        *config_.switchInputMethodBehavior ==
+            SwitchInputMethodBehavior::CommitDefault) {
+        chewing_cand_close(ctx);
+        if (chewing_buffer_Check(ctx)) {
+            // When not success, chewing_commit_preedit_buf will not change the
+            // output value. while chewing_handle_* will always update button
+            // result.
+            if (chewing_commit_preedit_buf(ctx) == 0) {
+                text.append(safeChewing_commit_String(ctx));
+            }
         }
     }
+
+    if (*config_.switchInputMethodBehavior ==
+        SwitchInputMethodBehavior::CommitPreedit) {
+        text.append(safeChewing_buffer_String(ctx));
+        text.append(safeChewing_bopomofo_String(ctx));
+    }
+    if (!text.empty()) {
+        event.inputContext()->commitString(text);
+    }
     doReset(event);
 }
 
